#ifndef LAPACK_DEFINES_H
#define LAPACK_DEFINES_H

// auto-generated by: 
// D:/bld/lapackpp_1639254729362/_build_env/Library/bin/cmake.exe 3.21.3
// host: fv-az59-768
//
// Definitions for:
// CXX   = 
//         C:/Program Files (x86)/Microsoft Visual Studio/2019/Enterprise/VC/Tools/MSVC/14.16.27023/bin/HostX64/x64/cl.exe
// LIBS  = 
//         D:/bld/lapackpp_1639254729362/_h_env/Library/lib/blas.lib
//         D:/bld/lapackpp_1639254729362/_h_env/Library/lib/lapack.lib;D:/bld/lapackpp_1639254729362/_h_env/Library/lib/blas.lib;-fortranlibs

#define LAPACK_VERSION 30800


#endif /* LAPACK_DEFINES_H */
